"""
The Utils and Measurements class.
"""

import sys
import numpy as np
import pandas as pd
from typing import List
from .household import Household


class Utilities:
    """
    Class containing a number of measurement functions.

    Todo:
        * Maybe remove this class?
    """

    def __init__(self):
        pass


class Measurements:
    """
    Class storing the segregation measurements per step.

    Args:
        model (CompassModel): CompassModel object.

    Attributes:
        all_measurements (list): list of all measurements.
        current (dict): counts the number of happy and unhappy agents.
        model (CompassModel): CompassModel object.
        agents (list): all the agents in the model
        vis_data (dict): dictionary for visualisation purposes.

    """

    def __init__(self, model):

        self.all_measurements: List[Object] = []
        self.model: "CompassModel" = model
        self.params: dict = model.params
        self.agents: dict[str, List[object]] = model.agents
        self.vis_data: dict = dict()

    def headers(self):
        """
        This function creates numpy arrays with the names of the columns for
        the household, neighbourhood and school data.
        """
        self.household_headers: np.ndarray = np.array(
            [
                "loc_x",
                "loc_y",
                "local_comp_0",
                "local_comp_1",
                "utility",
                "category",
                "id",
                "distance",
                "unit",
            ]
        )
        self.neighbourhood_headers: np.ndarray = np.array(
            ["comp_0", "comp_1", "utility", "satisfied", "distance", "unit"]
        )
        self.school_headers: np.ndarray = np.array(
            ["comp_0", "comp_1", "utility", "satisfied", "distance", "unit"]
        )

    def measurement_arrays(self) -> None:
        """
        Initialises the measurement arrays.
        """
        dtype = "float32"
        # Maximum steps and number of attributes per array
        max_steps = self.params["max_res_steps"] + 1 + self.params["max_school_steps"]
        self.headers()
        self.n_household_attrs: int = len(self.household_headers)
        self.n_neighbourhood_attrs: int = len(self.neighbourhood_headers)
        self.n_school_attrs: int = len(self.school_headers)
        self.temp_household: np.ndarray = np.zeros(self.n_household_attrs, dtype=dtype)

        # Determine the maximum shape of the arrays
        households_shape = (
            max_steps,
            self.params["n_households"],
            self.n_household_attrs,
        )
        neighbourhoods_shape = (
            self.params["max_res_steps"] + 1,
            self.params["n_neighbourhoods"],
            self.n_neighbourhood_attrs,
        )
        schools_shape = (
            self.params["max_school_steps"],
            self.params["n_schools"],
            self.n_school_attrs,
        )

        # Initialise empty arrays (datatype is important)
        self.households: np.ndarray = np.zeros(shape=households_shape, dtype=dtype)
        self.neighbourhoods: np.ndarray = np.zeros(
            shape=neighbourhoods_shape, dtype=dtype
        )
        self.schools: np.ndarray = np.zeros(shape=schools_shape, dtype=dtype)

    def household_data(self, residential: bool, time: int) -> None:
        """
        Gets the required data from all the households in the model.

        Args:
            residential (bool): True if we are in the residential process
            time (int): time step we are at

        Note:
            Double check neighbourhood and school indices (from unit) and the
            data
        """

        # Constant data
        if time == 0:
            for household in self.agents["households"]:
                self.households[:, household.idx, 0] = household.pos[0]
                self.households[:, household.idx, 1] = household.pos[1]
                self.households[:, household.idx, 2] = household.composition[0]
                self.households[:, household.idx, 3] = household.composition[1]
                self.households[:, household.idx, 5] = household.category
                self.households[:, household.idx, 6] = household.unique_id

        # Dynamic data
        if residential:
            self.households[time, :, 4] = Household._household_res_utility[:]
            self.households[time, :, 7] = Household._household_distance[:]
            for household in self.agents["households"]:
                self.households[
                    time, household.idx, 8
                ] = household.neighbourhood.unique_id
        else:
            self.households[time, :, 4] = Household._household_school_utility[:]
            self.households[time, :, 7] = Household._household_distance[:]
            self.households[time, :, 8] = Household._household_school_id[:]

    def neighbourhood_data(self, time: int) -> None:
        """
        Gets the missing data from all the neighbourhoods in the model.

        Args:
            time (int): time step we are at
        """
        for idx, neighbourhood in enumerate(self.agents["neighbourhoods"]):
            self.neighbourhoods[time, idx, :2] = neighbourhood.composition
            self.neighbourhoods[time, idx, 5] = neighbourhood.unique_id

    def school_data(self, time: int) -> None:
        """
        Gets the missing data from all the schools in the model.

        Args:
            time (int): time step we are at

        Note:
            time is different for schools compared to neighbourhoods!
        """

        for idx, school in enumerate(self.agents["schools"]):
            self.schools[time, idx, :2] = school.composition
            self.schools[time, idx, 5] = school.unique_id

    def end_step(self, residential: bool) -> None:
        """
        Perform end of cycle data collection. At the end of a cycle, every
        measurement is performed. The current measurements are only for
        households and are stored in a Numpy array with datatype uint16.
        """
        # Fill arrays with data
        self.residential = residential
        time = self.model.scheduler.get_time()
        if time == 0:
            self.measurement_arrays()
        self.household_data(residential, time)
        if residential:
            self.neighbourhood_data(time)
        else:
            self.school_data(self.model.scheduler.get_time("school") - 1)

    def get_last(self) -> object:
        """
        Returns:
            int: measurements from last performed step.
        """
        return self.all_measurements[-1]

    def get_bokeh_vis_data(self) -> object:
        """
        Stores model data in the correct format for the Bokeh visualisation.

        Returns:
            DataFrame of all the Bokeh visualisation data.

        Note:
            * For simplicity it is now in one Pandas DataFrame, for easy
            updating/filtering/selecting on the Bokeh server.
        """
        # Collect data from all the different agent types
        household_data = self.vis_household_data()
        school_data = self.vis_school_data(household_data)
        neighbourhood_data = self.vis_neighbourhood_data(household_data)
        system_data = self.vis_system_data(
            household_data, school_data, neighbourhood_data
        )
        vis_data = pd.concat(
            [household_data, school_data, neighbourhood_data, system_data],
            ignore_index=True,
        )
        # Incorporate the time step of the simulation
        vis_data["time"] = np.repeat(self.model.scheduler.get_time(), len(vis_data))
        return vis_data

    def empty_dataframe(
        self, columns: List[str] = None, n_rows: int = 0
    ) -> pd.DataFrame:
        """
        Creates an empty Pandas Dataframe

        Args:
            columns (list): list of column names in string format.
            n_rows (int): number of rows.

        Returns:
            Empty DataFrame for all the Bokeh visualisation data.
        """
        if columns is None:
            columns = []
        return pd.DataFrame(index=range(n_rows), columns=columns)

    def vis_composition_data(self, household: Household) -> List[np.ndarray]:
        """
        Extracts the composition data from the households.

        Args:
            household: a Household object
        """
        try:
            school_comp = household.students[0].school.composition
        except AttributeError:
            school_comp = household.new_composition_array()
        return [
            household.composition,
            household.neighbourhood.composition,
            school_comp,
            household.school_utility_comp,
        ]

    def vis_household_data(self) -> dict[str, object]:
        """
        Transforms the household data to be suitable for the Bokeh visualisation.

        Returns:
            DataFrame of all the household data.
        """

        # Grab the different times
        time = self.model.scheduler.get_time()
        res_time = self.model.scheduler.get_time("residential")

        households = self.agents["households"]
        columns = [
            "agent_type",
            "x",
            "y",
            "group0",
            "group1",
            "res_id",
            "res_utility",
            "res_happy",
            "school_id",
            "dist_school",
            "school_utility",
            "school_happy",
            "res_q5",
            "res_q95",
            "school_q5",
            "school_q95",
            "res_seg",
            "school_seg",
            "local_comp",
            "n_comp",
            "s_comp",
            "school_comp_utility",
        ]
        data = self.empty_dataframe(columns=columns, n_rows=len(households))

        # Save location and local composition per group type
        data["agent_type"] = "household"
        data["x"] = self.households[time, :, 0]
        data["y"] = self.households[time, :, 1]
        data["group0"] = (self.households[time, :, 5] == 0).astype(int)
        data["group1"] = (self.households[time, :, 5] == 1).astype(int)

        # Neighbourhood ID, current residential utility
        data["res_id"] = self.households[res_time, :, 8]
        data["res_utility"] = self.households[res_time, :, 4]
        data["res_happy"] = None

        composition_data = pd.DataFrame(
            [self.vis_composition_data(household) for household in households]
        )
        data[
            ["local_comp", "n_comp", "s_comp", "school_comp_utility"]
        ] = composition_data

        # Fill school data if applicable, set them to zero otherwise
        if not self.residential:
            data["school_id"] = (
                self.households[time, :, 8] - self.params["n_neighbourhoods"]
            ).astype(int)
            data["dist_school"] = self.households[time, :, 7]
            data["school_utility"] = self.households[time, :, 4]
            data["school_happy"] = None
        else:
            data["school_utility"] = 0
            data["dist_school"] = 0

        return data

    def vis_school_data(self, household_data: pd.DataFrame):
        """
        Gets the required data from all the schools in the model.

        Args:
            household_data (DataFrame): all the household data already gathered.

        Returns:
            DataFrame of all the school data.

        Note:
            School data is still calculated per household and not per student
        """
        schools = self.agents["schools"]
        data = self.empty_dataframe(columns=household_data.columns, n_rows=len(schools))

        for index, school in enumerate(schools):
            agent_type = "school"
            x, y = school.pos
            group0, group1 = school.composition.astype(int)
            res_id = None
            res_utility = None
            res_happy = None

            # School attributes

            # Subtract the number of neighbourhoods for the visualisation of
            # school composition plot.
            school_id = int(school.unique_id - self.params["n_neighbourhoods"])
            pupils = household_data[household_data.school_id == school_id]
            dist_school = pupils.dist_school.mean()
            school_utility = pupils.school_utility.mean()
            school_happy = pupils.school_happy.mean()
            school_comp_utility = pupils.school_comp_utility.mean()

            res_q5, res_q95, school_q5, school_q95, res_seg, school_seg = [None] * 6
            local_comp, n_comp, s_comp = [None] * 3
            s_comp = school.composition.astype(int)

            # Add data to the DataFrame
            data.iloc[index] = [
                agent_type,
                x,
                y,
                group0,
                group1,
                res_id,
                res_utility,
                res_happy,
                school_id,
                dist_school,
                school_utility,
                school_happy,
                res_q5,
                res_q95,
                school_q5,
                school_q95,
                res_seg,
                school_seg,
                local_comp,
                n_comp,
                s_comp,
                school_comp_utility,
            ]

        return data

    def vis_neighbourhood_data(self, household_data: pd.DataFrame):
        """
        Gets the required data from all the neighbourhood in the model.

        Args:
            household_data (DataFrame): all the household data already gathered.

        Returns:
            DataFrame of all the neighbourhood data.
        """
        neighbourhoods = self.agents["neighbourhoods"]
        data = self.empty_dataframe(
            columns=household_data.columns, n_rows=len(neighbourhoods)
        )

        for index, neighbourhood in enumerate(neighbourhoods):
            agent_type = "neighbourhood"
            group0, group1 = neighbourhood.composition.astype(int)
            res_id = index
            households = household_data[
                household_data.res_id == neighbourhood.unique_id
            ]
            res_utility = households.res_utility.mean()
            res_happy = households.res_happy.mean()
            x, y = neighbourhood.pos

            if neighbourhood.shape.type == "Polygon":
                x, y = neighbourhood.shape.exterior.coords.xy
            elif neighbourhood.shape.type == "MultiPolygon":
                x, y = neighbourhood.shape.convex_hull.exterior.coords.xy

            x, y = list(x), list(y)

            # School attributes
            school_id, dist_school, school_utility, school_happy = [None] * 4
            res_q5, res_q95, school_q5, school_q95, res_seg, school_seg = [None] * 6
            local_comp, n_comp, s_comp, school_comp_utility = [None] * 4

            # Add data to the DataFrame
            data.iloc[index] = [
                agent_type,
                x,
                y,
                group0,
                group1,
                res_id,
                res_utility,
                res_happy,
                school_id,
                dist_school,
                school_utility,
                school_happy,
                res_q5,
                res_q95,
                school_q5,
                school_q95,
                res_seg,
                school_seg,
                local_comp,
                n_comp,
                s_comp,
                school_comp_utility,
            ]
        return data

    def vis_system_data(
        self,
        household_data: pd.DataFrame,
        school_data: pd.DataFrame,
        neighbourhood_data: pd.DataFrame,
    ):
        """
        Gets the required data from the whole system.

        Args:
            household_data (DataFrame): all the household data already gathered.
            school_data (DataFrame): all the school data already gathered.
            neighbourhood_data (DataFrame): all the neighbourhood data already
                gathered.

        Returns:
            DataFrame of all the system data.
        """
        data = self.empty_dataframe(columns=household_data.columns, n_rows=1)
        agent_type = "system"
        x, y, group0, group1, res_id, school = 6 * [None]
        res_utility = household_data.res_utility.mean()
        res_happy = household_data.res_happy.mean()
        dist_school = household_data.dist_school.mean()
        school_utility = household_data.school_utility.mean()
        school_happy = household_data.school_happy.mean()

        res_q5 = household_data.res_utility.quantile(q=0.05)
        res_q95 = household_data.res_utility.quantile(q=0.95)
        school_q5 = household_data.school_utility.quantile(q=0.05)
        school_q95 = household_data.school_utility.quantile(q=0.95)
        res_seg = self.calculate_segregation(
            type="bounded_neighbourhood", index="Theil"
        )
        if self.residential:
            school_seg = 0
        else:
            school_seg = self.calculate_segregation(type="school", index="Theil")

        local_comp, n_comp, s_comp = [None] * 3
        school_comp_utility = household_data.school_comp_utility.mean()

        # Add data to the DataFrame
        data.iloc[0] = [
            agent_type,
            x,
            y,
            group0,
            group1,
            res_id,
            res_utility,
            res_happy,
            school,
            dist_school,
            school_utility,
            school_happy,
            res_q5,
            res_q95,
            school_q5,
            school_q95,
            res_seg,
            school_seg,
            local_comp,
            n_comp,
            s_comp,
            school_comp_utility,
        ]
        return data

    def export_data(self):
        """
        Export the data using numpy save.
        """

        if self.model.export:
            end_time = self.model.scheduler.get_time()
            res_end_time = self.model.scheduler.get_time("residential")
            if res_end_time == 0:
                res_end_time = 1
            school_end_time = self.model.scheduler.get_time("school")

            filename = self.model.params["filename"]
            if self.model.params["save_last_only"]:
                start = end_time - 1
                res_start = res_end_time - 1
                school_start = school_end_time - 1
                households = self.households[[res_start, start], :, :]

            else:
                res_start = 0
                school_start = 0
                households = self.households[:end_time, :, :]

            print("Saving data...")

            np.savez(
                filename,
                households=households,
                chosen_indices=self.model.chosen_indices,
                households_headers=self.household_headers,
                neighbourhoods=self.neighbourhoods[res_start:res_end_time, :, :],
                neighbourhoods_headers=self.neighbourhood_headers,
                schools=self.schools[school_start:school_end_time, :, :],
                schools_headers=self.school_headers,
                params=self.params,
            )
            print("Data saved!")

    def calculate_segregation(
        self, type: str = "school", index: str = "Theil", per_location: bool = False
    ):
        """
        Calculate segregation index for the whole system.

        Args:
            type (str): calculates 'school' (defaul), 'bounded_neighbourhood'
                (Neighbhourhood) or 'local_neighbourhood' (Moore) segregation.
            index (str): 'Theil' Entropy based segregation index.
            per_location (bool): Default is False, but if True it returns the
                decomposed index.

        Returns:
            float: segregation index for the whole system.
            list: if per_location is set to True.
        """
        if index == "Theil":
            return self.calculate_theil(type, per_location)
        else:
            print("Segregation index not supported")
            exit(1)

    def calculate_theil(self, type: str, per_location: bool = False):
        """
        Calculate Theil's information index.

        Args:
            type (str): calculates 'school' (defaul), 'bounded_neighbourhood'
                (Neighbhourhood) or 'local_neighbourhood' (Moore) segregation.
            index (str): 'Theil' Entropy based segregation index.
            per_location (bool): Default is False, but if True it returns the
                decomposed index.

        Returns:
            float: segregation index for the whole system.
            list: if per_location is set to True.

        Note:
            Only works for first category in self.params["group_categories"].

        Todo:
            Decide which notation to use
        """

        # Check which composition to use
        if type == "bounded_neighbourhood":
            agents = self.agents["neighbourhoods"]
        elif type == "local_neighbourhood":
            agents = self.agents["households"]
        elif type == "school":
            agents = self.agents["schools"]
        else:
            print("Calculation of Theil's information index not supported.")
            sys.exit(1)

        pi_m = self.model.global_composition / self.model.global_composition.sum()

        local_compositions = np.empty((len(agents), len(pi_m)))
        nr_of_agents = np.empty(len(agents))
        for i, agent in enumerate(agents):
            nr_of_agents[i] = np.sum(agent.composition)
            # TODO: is there a way to move the check out of the for loop?
            if nr_of_agents[i] < 1:
                local_compositions[i] = agent.composition
            else:
                local_compositions[i] = agent.composition / nr_of_agents[i]

        total_agents = np.sum(nr_of_agents)

        pi_jm = local_compositions
        t_j = nr_of_agents
        T = total_agents
        r_jm = pi_jm / pi_m

        global_entropy = -np.sum(pi_m * np.log(pi_m))
        E = global_entropy
        log_r_jm = np.nan_to_num(np.log(r_jm))

        # Full sum if combined, leave as array if per location
        if per_location:
            H = np.sum((t_j / (T * E)) * (pi_jm * log_r_jm).T, axis=0)
        else:
            H = np.sum((t_j / (T * E)) * (pi_jm * log_r_jm).T, axis=None)
        theil = H
        return theil



"""
The Student class.
"""
from typing import List, Iterable


class Student():
    """
    Student object that is enrolled into school objects and a Household. Used
    for measuring segregation in schools and neighbourhoods.

    Args:
        unique_id (int): unique identifier of the agent.
        household (Household): Household object.
        groups (list): containing all group types.

    Attributes:
        unique_id (int): unique identifier of the agent.
        household (Household): Household object.
        groups (list): containing all group types.
        school_preference ():
        school (School): School object that the student is enrolled in.
        satisfied (bool): True if the student is satisfied with current school.
        school_history (list): all the school objects the student has attended.

    """

    _total_students = 0

    def __init__(self, unique_id: int, household: 'Household'):

        self.idx = Student._total_students  # type: int
        Student._total_students += 1

        self.school = None  # type: 'School'
        self.school_history = []  # type: List['School']
        self.unique_id = unique_id  # type: int
        self.household = household  # type: 'Household'
        self.school_preference = None  # type: Iterable[int]
        # Student does not inherit from BaseAgent, so increment here.
        self.household.model.increment_agent_count()

    def __repr__(self) -> str:
        """
        Returns:
            str: representing the unique identifier of the agent.
        """
        return f"<Student object with unique_id:{self.unique_id}>"

    def set_school_preference(self, ranking: Iterable[int]) -> None:
        """
        Sets school preference to a given ranking.

        Args:
            ranking (list): a ranking of School objects.
        """
        self.school_preference = ranking

    def new_school(self, school: 'School') -> None:
        """
        Enrolls the student in a new school.

        Args:
            school (School): a School object.
        """
        if self.school:
            self.school.remove_student(self)
        self.school = school
        self.school_history.append(school)
        school.add_student(self)

    def get_school_id(self) -> int:
        """
        Returns:
            str: unique id of the school the student is enrolled in.
        """
        if self.school:
            return self.school.unique_id

        return -1



from re import I
import time
import logging
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from os.path import dirname, join
from tornado import gen
import numpy as np

# Bokeh imports
from bokeh.plotting import figure
from bokeh.server.server import Server
from bokeh.transform import linear_cmap
from bokeh.application import Application
from bokeh.tile_providers import get_provider
from bokeh.layouts import row, column, gridplot
from bokeh.document import without_document_lock
from bokeh.application.handlers.function import FunctionHandler
from bokeh.models import (ColumnDataSource, Button, Select, Slider, CDSView,
                          BooleanFilter, HoverTool, Dropdown, Div, Band,
                          WheelZoomTool)

from .parameters import FLAGS
from .model import CompassModel


class BokehServer():
    """
    Class to run the Bokeh server for visualisation.

    Args:
        doc: a Bokeh Document instance

    Attributes:
        grid_plot: initiates the plot for the grid
        line_plot: initiates one line plot
        update: updates the plots with every time step
        visualise_model: starts the whole visualisation
        grid_values: obtains all the grid values from Mesa.grid
    """

    def __init__(self, doc):

        # This is important! Save curdoc() to make sure all threads
        # see the same document.
        self.doc = doc

        # Initialise a CompassModel model with the default setting.
        self.model = CompassModel(**vars(FLAGS))
        self.params = self.model.params
        self.res_ended = False
        self.school_ended = False
        self.executor = ThreadPoolExecutor(max_workers=2)

        # Initialise layout and populate the plots with the initial
        # configuration
        self.reset_data()
        self.layout()
        print('Visualisation running...')

    def agent_filter(self, source, agent_type):
        """
        This function creates a boolean agent filter view for the Bokeh
        visualisation instance.

        Args:
            source (CDS): ColumnDataSource from Bokeh.
            agent_type (str): the agent type to be filtered.

        Returns:
            view: the view that can be used in plots.
        """

        by_type = [True if agent == agent_type else False \
                    for agent in source.data['agent_type']]
        view = CDSView(source=source, filters=[BooleanFilter(by_type)])
        return view

    def customise_grid(self):
        """
        Grid customisation; agent colors, neighbourhood boundaries and schools.

        Todo:
            size should come from a parameter, as they do not scale properly
            with grid size.
        """

        # Agent colours, agent tooltips and grid initialisation
        mapper = linear_cmap(field_name='group1',
                             palette=['blue', 'red'],
                             low=0,
                             high=1)
        TOOLTIPS = [("Residential utility", "@res_utility"),
                    ('Local composition', '@local_comp'),
                    ('Neighbourhood composition', '@n_comp'),
                    ("School utility", "@school_utility"),
                    ('Distance', '@dist_school'),
                    ('School composition', '@s_comp'),
                    ('School composition utility', '@school_comp_utility')]
        hover = HoverTool(names=["households", "schools"], tooltips=TOOLTIPS)
        self.grid = figure(x_range=(self.model.grid.x_min - 1,
                                    self.model.grid.x_max),
                           y_range=(self.model.grid.y_min - 1,
                                    self.model.grid.y_max),
                           tools=[hover, 'tap', 'pan',
                                  WheelZoomTool()],
                           tooltips=TOOLTIPS)

        # Set WheelZoomTool active by default if not lattice
        if self.params['case'].lower() != 'lattice':
            self.grid.toolbar.active_scroll = self.grid.select_one(
                WheelZoomTool)

        # Add a background map using OpenStreetMap (Google Maps is too
        # computationally expensive and cannot zoom properly)
        self.grid.add_tile(get_provider('OSM'))

        self.grid.axis.visible = False
        self.grid.grid.visible = False
        # Function to highlight all households that are currently enrolled in
        # the same school.
        self.source.selected.on_change("indices", self.select_households)

        # Plot households
        self.grid.circle(x="x",
                         y="y",
                         size=5,
                         view=self.household_view,
                         source=self.source,
                         fill_color=mapper,
                         line_color='black',
                         alpha=0.8,
                         nonselection_fill_alpha=0.2,
                         selection_fill_alpha=1,
                         name='households')

        # Plot schools
        self.grid.circle(x="x",
                         y="y",
                         size=7,
                         source=self.source,
                         view=self.school_view,
                         fill_color='yellow',
                         line_color='black',
                         name='schools')

        # Plot neighbourhoods
        self.grid.patches('x',
                          'y',
                          source=self.source,
                          view=self.neighbourhood_view,
                          fill_color=None,
                          line_color="black",
                          line_width=2,
                          hover_alpha=0,
                          hover_line_color=None,
                          name='neighbourhoods',
                          selection_fill_alpha=0)

    def init_grid_plot(self):
        """
        Initiates the grid plot for the visualisation.
        """

        # Create filters to plot households and schools sequentially
        self.household_view = self.agent_filter(self.source, 'household')
        self.school_view = self.agent_filter(self.source, 'school')
        self.neighbourhood_view = self.agent_filter(self.source,
                                                    'neighbourhood')
        self.customise_grid()

    def select_households(self, attr, old, new):
        """
        This function selects all households that are in the same school.

        Args:
            attr: -
            old: -
            new (list): indices that are clicked (can be multiple)
        """
        index = new[0]
        school_id = self.data.iloc[index].school_id
        # List all agents that have the same school id
        same_school = self.data[self.data.school_id == school_id].index
        self.source.selected.indices = list(same_school)

    def dropdown_select(self, event):
        """
        This function creates a dropdown of all schools.

        Args:
            event (str): the selected school
        """

        school_id = int(event.item)
        same_school = self.data[self.data.school_id == school_id].index
        self.source.selected.indices = list(same_school)

    def school_dropdown_func(self, width=100):
        """
        Event listener for the dropdown of all schools.
        """
        schools = self.data[self.data.agent_type == 'school']
        school_locs = [str(num) for num in range(len(schools))]
        self.school_dropdown = Dropdown(label='Select School',
                                        menu=school_locs,
                                        width=width)
        self.school_dropdown.on_click(self.dropdown_select)

    def init_line_plot(self, width=200, height=200, mode='fixed'):
        """
        Initiates the line plot for the server.
        """

        # Create a ColumnDataSource that can be updated at every step.
        TOOLTIPS = [
            ("Residential utility", "@res_utility"),
            ("Residential segregation", "@res_seg"),
            ("School utility", "@school_utility"),
            ("School segregation", "@school_seg"),
        ]
        self.plot = figure(tooltips=TOOLTIPS,
                           y_range=(0, 1),
                           plot_width=width,
                           sizing_mode=mode,
                           title="Neighbourhood/school utility/segregation")

        plot_pars = {
            'Residential utility': {
                'y': 'res_utility',
                'color': 'green',
                'lower': 'res_q5',
                'upper': 'res_q95'
            },
            'Residential segregation': {
                'y': 'res_seg',
                'color': 'blue',
                'lower': None,
                'upper': None
            },
            'School utility': {
                'y': 'school_utility',
                'color': 'orange',
                'lower': 'school_q5',
                'upper': 'school_q95'
            },
            'School segregation': {
                'y': 'school_seg',
                'color': 'purple',
                'lower': None,
                'upper': None
            }
        }

        # Plot lines, markers and bands (for utility only)
        for label in plot_pars.keys():
            y = plot_pars[label]['y']
            color = plot_pars[label]['color']
            lower = plot_pars[label]['lower']
            upper = plot_pars[label]['upper']
            self.plot.line(x='time',
                           y=y,
                           source=self.line_source,
                           line_width=2,
                           color=color,
                           legend_label=label)
            self.plot.circle(x='time',
                             y=y,
                             source=self.line_source,
                             size=5,
                             color=color,
                             legend_label=label)

            if lower is not None:
                band = Band(base='time',
                            lower=lower,
                            upper=upper,
                            source=self.line_source,
                            fill_alpha=0.2,
                            fill_color=color)
            self.plot.add_layout(band)

        self.plot.legend.location = 'top_left'

    def composition_data(self, agent_type='school'):
        """
        Calculates the composition data for schools or neighbourhoods.

        Args:
            agent_type (str): either 'school' or 'neighbourhood'

        Returns:
            DataFrame of fraction blues and reds per agent.
        """

        # Loop over all schools to get the locations and compositions
        cols = ['group0', 'group1', 'dist_school']
        data = self.data[self.data.agent_type == agent_type][cols]

        if agent_type == 'household':
            blues = data[data['group0'] == 1]['dist_school'].values
            reds = data[data['group1'] == 1]['dist_school'].values
            return {'blues': blues, 'reds': reds}
        else:
            totals = data.sum(axis=1).replace(0, 1).values
            fractions = data.div(totals, axis=0)
            blues = fractions['group0'].values
            reds = fractions['group1'].values
            return {'blues': blues, 'reds': reds}

    def init_school_composition_plot(self,
                                     width=200,
                                     height=200,
                                     mode='fixed'):
        """
        Initiates the school composition plot.
        """

        self.school_composition_quads = {}
        self.school_composition_plot = figure(title="School composition",
                                              x_range=(0, 1),
                                              plot_width=width,
                                              sizing_mode=mode)

        fractions = self.composition_data(agent_type='school')
        for group in fractions.keys():

            hist, edges = np.histogram(fractions[group], density=True, bins=20)
            self.school_composition_quads[
                group] = self.school_composition_plot.quad(
                    top=hist,
                    bottom=0,
                    left=edges[:-1],
                    right=edges[1:],
                    fill_color=group[:-1],
                    line_color="white",
                    alpha=0.7,
                    legend_label=group)

    def init_neighbourhood_composition_plot(self,
                                            width=200,
                                            height=200,
                                            mode='fixed'):
        """
        Initiates the neighbourhood composition plot.
        """
        self.neighbourhood_composition_quads = {}
        self.neighbourhood_composition_plot = figure(
            title="Neighbourhood composition",
            x_range=(0, 1),
            plot_width=width,
            sizing_mode=mode)

        fractions = self.composition_data(agent_type='neighbourhood')
        for group in fractions.keys():

            hist, edges = np.histogram(fractions[group], density=True, bins=20)
            self.neighbourhood_composition_quads[group] = \
                self.neighbourhood_composition_plot.quad(
                top=hist, bottom=0, left=edges[:-1], right=edges[1:],
                fill_color=group[:-1], line_color="white", alpha=0.7,
                legend_label=group)

    def init_distribution_plot(self, width=200, height=200, mode='fixed'):
        """
        Initiates the distribution plots for residential and school utility.
        """

        self.distribution_plot = figure(title="Utility distributions",
                                        x_range=(0, 1),
                                        plot_width=width,
                                        sizing_mode=mode)

        hist_data = self.data[self.data.agent_type == 'household'][[
            'res_utility', 'school_utility'
        ]]

        # Residential utility
        hist, edges = np.histogram(hist_data.res_utility,
                                   density=True,
                                   bins=50)
        self.res_quads = self.distribution_plot.quad(
            top=hist,
            bottom=0,
            left=edges[:-1],
            right=edges[1:],
            fill_color="green",
            line_color="white",
            alpha=0.7,
            legend_label='Residential utility')

        # School utility
        if not self.residential:
            hist, edges = np.histogram(hist_data.school_utility,
                                       density=True,
                                       bins=50)
        # Do not show school utilities yet as they are all zero!
        hist = np.zeros(len(hist))
        self.school_quads = self.distribution_plot.quad(
            top=hist,
            bottom=0,
            left=edges[:-1],
            right=edges[1:],
            fill_color="orange",
            line_color="white",
            alpha=0.7,
            legend_label='School utility')

    def init_distance_plot(self, width=200, height=200, mode='fixed'):
        """
        Initiates the distance plot for the school choice process.
        """
        """
        Initiates the neighbourhood composition plot.
        """
        self.distance_quads = {}
        self.distance_plot = figure(title="Distance utility",
                                    x_range=(0, 1),
                                    plot_width=width,
                                    sizing_mode=mode)

        fractions = self.composition_data(agent_type='household')
        for group in fractions.keys():

            hist, edges = np.histogram(fractions[group], density=True, bins=20)
            self.distance_quads[group] = \
                self.distance_plot.quad(
                top=hist, bottom=0, left=edges[:-1], right=edges[1:],
                fill_color=group[:-1], line_color="white", alpha=0.7,
                legend_label=group)

    def update_filters(self):
        """
        Updates the view filters for households, schools and neighbourhoods as
        they can change when reset is clicked (i.e., new model instance).
        """

        # Update household filter
        household_filter = [True if agent == 'household' else False for agent \
            in self.source.data['agent_type']]
        self.household_view.filters[0] = BooleanFilter(household_filter)

        # Update neighbourhood filter
        neighbourhood_filter = [True if agent == 'neighbourhood' else False for\
            agent in self.source.data['agent_type']]
        self.neighbourhood_view.filters[0] = BooleanFilter(
            neighbourhood_filter)

        # Update school filter
        school_filter = [True if agent == 'school' else False for agent in \
            self.source.data['agent_type']]
        self.school_view.filters[0] = BooleanFilter(school_filter)

    def reset_data(self):
        """
        Resets the data, could be the initial reset (new sources need to be
        created) or a subsequent one (only update the data).
        """

        # Callback object to run, step and reset the model properly.
        self.residential = True
        self.callback_obj = None
        self.data = self.model.measurements.get_bokeh_vis_data()

        # Check if it's the initial reset (create new sources) or a reset button
        # click (update .data only)
        try:
            self.source.data = self.data
            self.line_source.data = self.data[self.data.agent_type == 'system']
            self.update_filters()
            self.update_data()

        # Reset is clicked --> update .data only
        except AttributeError:
            self.source = ColumnDataSource(self.data)
            self.line_source = ColumnDataSource(
                self.data[self.data.agent_type == 'system'])

    def update_data(self):
        """
        Updates all data sources.
        """

        # Update all plots in the figure
        self.data = self.model.measurements.get_bokeh_vis_data()
        self.source.stream(self.data, len(self.data))
        self.line_source.stream(self.data[self.data.agent_type == 'system'])
        self.school_dropdown_func()

        # Update the utility histograms
        self.update_histograms()

        # Update the composition histograms
        for quads in [
                self.neighbourhood_composition_quads,
                self.school_composition_quads, self.distance_quads
        ]:

            # Grab the new data
            if quads == self.neighbourhood_composition_quads:
                hist_data = self.composition_data(agent_type='neighbourhood')
            elif quads == self.school_composition_quads:
                hist_data = self.composition_data(agent_type='school')
            else:
                hist_data = self.composition_data(agent_type='household')

            # Update the bars and edges
            for group in hist_data.keys():

                hist, edges = np.histogram(hist_data[group],
                                           density=True,
                                           bins=20)

                # Update histogram
                quads[group].data_source.data['top'] = hist
                quads[group].data_source.data['left'] = edges[:-1]
                quads[group].data_source.data['right'] = edges[1:]

    def update_histograms(self):

        hist_data = self.data[self.data.agent_type == 'household'][[
            'res_utility', 'school_utility'
        ]]

        hist, edges = np.histogram(hist_data.res_utility,
                                   density=True,
                                   bins=50)
        self.res_quads.data_source.data['top'] = hist
        self.res_quads.data_source.data['left'] = edges[:-1]
        self.res_quads.data_source.data['right'] = edges[1:]

        # Only start to show school utilities when the residential process is
        # finished
        if not self.residential:
            hist, edges = np.histogram(hist_data.school_utility,
                                       density=True,
                                       bins=50)
        else:
            hist = np.zeros(len(hist))

        self.school_quads.data_source.data['top'] = hist
        self.school_quads.data_source.data['left'] = edges[:-1]
        self.school_quads.data_source.data['right'] = edges[1:]

    def blocking_task(self):
        time.sleep(0.1)

    @without_document_lock
    @gen.coroutine
    def unlocked_task(self):
        """
        Needed to make sure that if the reset button is clicked it can go
        inbetween events, otherwise it can be quite slow.
        """
        yield self.executor.submit(self.blocking_task)
        self.doc.add_next_tick_callback(partial(self.step_button))

    def run_button(self):
        """
        Handles the run button clicks, coloring and starts the simulation.
        """
        if self.run.label == 'Run':
            self.run.label = 'Stop'
            self.run.button_type = 'danger'
            self.callback_obj = self.doc.add_periodic_callback(
                self.unlocked_task, 500)

        else:
            self.run.label = 'Run'
            self.run.button_type = 'success'
            self.doc.remove_periodic_callback(self.callback_obj)

    def step_button(self):
        """
        Checks which process need to be stepped and execute the step. The
        simulate function of the Model instance cannot be used as we need to
        visualise every step.
        """

        # Plus 1 because the first time step is already plotted
        max_res_steps = self.params['max_res_steps']
        max_school_steps = self.params['max_school_steps']
        school_time = self.model.scheduler.get_time('school')
        residential_time = self.model.scheduler.get_time('residential') + 1

        # If residential is not converged yet or below max steps, do a step
        if (residential_time <= max_res_steps and not self.res_ended):
            self.res_ended = self.model.convergence_check()
            if not self.res_ended:
                self.model.step(residential=True)
            else:
                print('Residential process ended.')

        # Initial school step needs to be executed
        elif school_time == 0:
            self.residential = False
            self.model.step(residential=False, initial_schools=True)

        # Normal school steps
        elif (school_time < max_school_steps and not self.school_ended):
            self.school_ended = self.model.convergence_check()
            if not self.school_ended:
                self.model.step(residential=False, initial_schools=False)
            else:
                print('School process ended.')

        # Both processes are done/converged
        else:
            self.run_button()
            return

        self.update_data()

    def reset_button(self):
        """
        Resets the model and takes the (possible) new parameter values into
        account.
        """

        # Update the parameter values and start a new model
        self.time_series = []
        self.res_ended = False
        self.school_ended = False
        self.model = CompassModel(self.update_pars())

        # Stop the model when it is still running while reset is clicked.
        if self.run.label == 'Stop' and self.callback_obj is not None:
            self.doc.remove_periodic_callback(self.callback_obj)
            self.run.label = 'Run'
            self.run.button_type = 'success'

        self.reset_data()
        self.doc.clear()
        self.layout()

    def layout(self):
        """
        Sets up the whole layout; widgets and all plots.
        """

        # Initialise all plots and widgets
        widgets = self.widgets(width=200)

        plot_width = 500
        sizing_mode = 'stretch_height'
        self.init_grid_plot()
        self.init_line_plot(width=plot_width, mode=sizing_mode)
        self.init_distribution_plot(width=plot_width, mode=sizing_mode)
        self.init_school_composition_plot(width=plot_width, mode=sizing_mode)
        self.init_neighbourhood_composition_plot(width=plot_width,
                                                 mode=sizing_mode)
        self.init_distance_plot(width=plot_width, mode=sizing_mode)

        # Row with widgets
        if self.params['case'].lower() == 'lattice':
            width = 420
            split = int(len(widgets) / 2.) + 1
            widget_row = row(
                [column(widgets[:split]),
                 column(widgets[split:])],
                width=width)
        else:
            width = 210
            widget_row = column(widgets, width=width)

        desc = Div(text=open(join(dirname(__file__),
                                  "description.html")).read(),
                   margin=0)
        # Column with all the controls and description
        first_col = column(widget_row, width=width, sizing_mode='fixed')

        # Column with the grid/map
        second_col = column([
            desc,
            row(self.buttons(), sizing_mode='stretch_width'),
            row(self.grid, sizing_mode='stretch_width')
        ],
                            sizing_mode='stretch_width')

        # Column with the plots
        third_col = column([
            self.plot, self.distribution_plot, self.distance_plot,
            self.school_composition_plot, self.neighbourhood_composition_plot
        ])

        # second_third = row([second_col, third_col], sizing_mode='scale_width')

        vis_layout = gridplot([[first_col, second_col, third_col]],
                              toolbar_location=None)

        self.doc.add_root(vis_layout)
        self.doc.title = "COMPASS"

    def buttons(self, width=100):
        self.run = Button(label="Run", button_type='success', height=32)
        self.run.on_click(self.run_button)
        self.step = Button(label="Step", button_type='primary', height=32)
        self.step.on_click(self.step_button)
        self.reset = Button(label="Reset", button_type='warning', height=32)
        self.reset.on_click(self.reset_button)
        buttons = [self.run, self.step, self.reset]
        return buttons

    def widgets(self, width=100):
        """
        Hardcodes all widgets.
        """

        header_size = '<h3>'

        # Simulation
        self.sleep = Slider(start=0,
                            end=1,
                            value=0.1,
                            step=0.1,
                            title='Time between steps',
                            width=width)
        self.max_move_fraction = Slider(start=0,
                                        end=1,
                                        value=self.params['max_move_fraction'],
                                        step=.05,
                                        title="Fraction moving",
                                        width=width)
        self.max_res_steps = Slider(start=0,
                                    end=1000,
                                    value=self.params['max_res_steps'],
                                    step=10,
                                    title="Residential steps",
                                    width=width)
        self.max_school_steps = Slider(start=0,
                                       end=1000,
                                       value=self.params['max_school_steps'],
                                       step=10,
                                       title="School steps",
                                       width=width)
        self.threshold = Select(title="Convergence threshold",
                                options=['0.001', '0.005', '0.01', '0.02'],
                                value=str(self.params['conv_threshold']),
                                width=width)
        self.homophily_std = Select(title="Std. of optimal fraction",
                                    options=[str(x / 100.) for x in range(11)],
                                    value=str(self.params['homophily_std']),
                                    width=width)
        self.temperature = Slider(start=0,
                                  end=500,
                                  step=1,
                                  title="Temperature",
                                  value=self.params['temperature'],
                                  width=width)
        self.window_size = Slider(start=10,
                                  end=50,
                                  value=self.params['window_size'],
                                  step=10,
                                  title="Convergence window size",
                                  width=width)
        self.case = Select(title='Case',
                           options=[
                               'Lattice',
                               'Amsterdam',
                           ],
                           value=str(self.params['case']),
                           width=width)

        if self.params['random_residential'] == 0:
            random_residential = 'False'
        else:
            random_residential = 'True'
        self.random_residential = Select(title="Random residential",
                                         options=['True', 'False'],
                                         value=random_residential,
                                         width=width)

        text = header_size + 'Simulation' + header_size
        simulation_div = Div(text=text, width=width)
        simulation = [
            simulation_div, self.max_move_fraction, self.max_res_steps,
            self.max_school_steps, self.threshold, self.homophily_std,
            self.window_size, self.temperature, self.case,
            self.random_residential
        ]

        if self.params['case'].lower() != 'lattice':
            text = header_size + 'Parameters' + header_size
            parameter_div = Div(text=text, width=width)
            simulation = [
                parameter_div, self.max_move_fraction, self.max_school_steps,
                self.threshold, self.homophily_std, self.window_size,
                self.temperature, self.case, self.random_residential
            ]

        # Grid
        self.size = Select(title="Size",
                           options=[str(x * 10) for x in range(1, 16)],
                           value=str(self.params['width']),
                           width=width)
        self.n_neighbourhoods = Select(
            title="Number of neighbourhoods",
            options=[str(x**2) for x in range(1, 16)],
            value=str(self.params['n_neighbourhoods']),
            width=width)
        self.n_schools = Select(title="Number of schools",
                                options=[str(x**2) for x in range(1, 11)],
                                value=str(self.params['n_schools']),
                                width=width)
        self.schools_placement = Select(
            title="School placement",
            options=['evenly_spaced', 'random', 'random_per_neighbourhood'],
            value=str(self.params['schools_placement']),
            width=width)
        self.household_density = Slider(start=0,
                                        end=1,
                                        value=self.params['household_density'],
                                        step=.05,
                                        title="Density",
                                        width=width)
        self.group_dist = Slider(start=0,
                                 end=1,
                                 value=self.params['group_dist'][0][0],
                                 step=.05,
                                 title="Share blue",
                                 width=width)
        self.torus = Select(title="Torus",
                            options=['True', 'False'],
                            value=str(self.params['torus']),
                            width=width)

        if self.params['scheduling'] == 0:
            scheduling = 'replacement'
        else:
            scheduling = 'without_replacement'
        self.scheduling = Select(
            title="Scheduling method",
            options=['replacement', 'without_replacement'],
            value=scheduling,
            width=width)

        text = header_size + 'Environment' + header_size
        environment_div = Div(text=text, width=width)
        grid = [
            environment_div, self.size, self.n_neighbourhoods, self.n_schools,
            self.household_density, self.group_dist, self.scheduling,
            self.torus, self.schools_placement
        ]

        if self.params['case'].lower() != 'lattice':
            grid = [self.scheduling]

        # School
        self.school_capacity = Slider(start=0,
                                      end=3,
                                      value=self.params['school_capacity'],
                                      step=.1,
                                      title="School capacity",
                                      width=width)
        self.school_dropdown_func(width=width)
        text = header_size + 'Schools' + header_size
        school_div = Div(text=text, width=width)
        school = [school_div, self.school_capacity, self.school_dropdown]

        if self.params['case'].lower() != 'lattice':
            school = [self.school_capacity, self.school_dropdown]

        # Household
        self.alpha = Slider(start=0,
                            end=1,
                            value=self.params['alpha'],
                            step=.05,
                            title="Alpha",
                            width=width)
        self.utility_at_max = Slider(start=0,
                                     end=1,
                                     value=self.params['utility_at_max'][0][0],
                                     step=.05,
                                     title="Utility at homogeneity",
                                     width=width)
        self.optimal_fraction = Slider(
            start=0,
            end=1,
            value=self.params['optimal_fraction'][0][0],
            step=.05,
            title="Optimal fraction",
            width=width)
        self.radius = Select(title="Radius",
                             options=[str(x) for x in range(1, 11)],
                             value=str(self.params['radius']),
                             width=width)
        self.neighbourhood_mixture = Slider(
            start=0,
            end=1,
            value=self.params['neighbourhood_mixture'],
            step=.05,
            title="Neighbourhood Mixture",
            width=width)
        self.num_considered = Slider(start=1,
                                     end=10,
                                     value=self.params['num_considered'],
                                     step=1,
                                     title="Considered spots",
                                     width=width)
        self.ranking_method = Select(title="Ranking method",
                                     options=['highest', 'proportional'],
                                     value=self.params['ranking_method'],
                                     width=width)
        self.p = Slider(start=500,
                        end=10000,
                        value=self.params['p'],
                        step=500,
                        title="Location of sigmoid",
                        width=width)
        self.q = Slider(start=1,
                        end=10,
                        value=self.params['q'],
                        step=1,
                        title="Slope of sigmoid",
                        width=width)

        text = header_size + 'Households' + header_size
        household_div = Div(text=text, width=width)
        household = [
            household_div, self.alpha, self.utility_at_max,
            self.optimal_fraction, self.radius, self.neighbourhood_mixture,
            self.num_considered, self.ranking_method, self.p, self.q
        ]

        if self.params['case'].lower() != 'lattice':
            household = [
                self.alpha, self.utility_at_max, self.optimal_fraction,
                self.ranking_method, self.p, self.q
            ]

        widgets = simulation + grid + school + household

        return widgets

    def update_pars(self):
        """
        Updates the parameters.
        """

        self.params['alpha'] = self.alpha.value
        self.params['max_move_fraction'] = self.max_move_fraction.value
        self.params['max_school_steps'] = self.max_school_steps.value
        self.params['conv_threshold'] = float(self.threshold.value)
        self.params['utility_at_max'] = [[
            self.utility_at_max.value, self.utility_at_max.value
        ]]
        self.params['optimal_fraction'] = [[
            self.optimal_fraction.value, self.optimal_fraction.value
        ]]
        self.params['school_capacity'] = self.school_capacity.value
        self.params['neighbourhood_mixture'] = self.neighbourhood_mixture.value
        self.params['temperature'] = self.temperature.value
        self.params['homophily_std'] = float(self.homophily_std.value)
        self.params['window_size'] = int(self.window_size.value)
        self.params['num_considered'] = int(self.num_considered.value)
        self.params['ranking_method'] = self.ranking_method.value
        self.params['case'] = self.case.value
        self.params['p'] = self.p.value
        self.params['q'] = self.q.value

        if self.random_residential.value == 'False':
            self.params['random_residential'] = 0
        else:
            self.params['random_residential'] = 1

        if self.scheduling.value == 'replacement':
            self.params['scheduling'] = 0
        else:
            self.params['scheduling'] = 1

        if self.params['case'].lower() != 'lattice':
            self.params['max_res_steps'] = 0
            return self.params

        self.params['width'] = int(self.size.value)
        self.params['height'] = int(self.size.value)
        self.grid.x_range.end = self.params['width']
        self.grid.y_range.end = self.params['height']
        self.params['torus'] = eval(self.torus.value)
        self.params['household_density'] = self.household_density.value
        self.params['max_res_steps'] = self.max_res_steps.value
        self.params['n_neighbourhoods'] = int(self.n_neighbourhoods.value)
        self.params['n_schools'] = int(self.n_schools.value)
        self.params['schools_placement'] = self.schools_placement.value
        self.params['group_dist'] = [[
            self.group_dist.value, 1 - self.group_dist.value
        ]]

        return self.params

    def run_visualisation(self):

        apps = {'/': Application(FunctionHandler(BokehServer))}
        server = Server(apps, port=5004)
        # To avoid bokeh's logger spamming
        log = logging.getLogger('bokeh')
        log.setLevel('ERROR')
        server.io_loop.add_callback(server.show, "/")
        server.io_loop.start()



"""
The Neighbourhood class.
"""
import numpy as np
from typing import List, ClassVar
from .agents_base import BaseAgent


class Neighbourhood(BaseAgent):
    """
    The Neighbourhood class.

    Args:
        unique_id (int): unique identifier of the agent.
        pos (tuple): (x,y) coordinates of the agent in the 2D-grid.
        model (CompassModel): CompassModel object.
        params (Argparser): containing all parameter values.

    Attributes:
        unique_id (int): unique identifier of the agent.
        pos (tuple): (x,y) coordinates of the agent in the 2D-grid.
        model (CompassModel): CompassModel object.
        params (Argparser): containing all parameter values.
        total (int): the number of Households in the neighbourhood.
        households (list): all the households living in the neighbourhood.
        composition (array): the sum of the attribute arrays of all Households
            belonging to this neighbourhood.
    """

    _total_neighbourhoods: ClassVar[int] = 0
    __slots__ = ["idx"]

    def __init__(
            self,
            unique_id: int,
            pos: tuple[float, float],
            shape,  # FIXME: type?
            model: object,
            params: dict
            ):

        super().__init__(unique_id, pos, model, params)

        self.idx: int = Neighbourhood._total_neighbourhoods
        Neighbourhood._total_neighbourhoods += 1

        self.total: int  = 0
        self.shape = shape
        self.households: List['Household'] = []
        self.composition: np.ndarray = self.new_composition_array()

    def __repr__(self) -> str:
        """
        Returns:
            str: representing the unique identifier of the agent.
        """
        return f"<Neighbourhood object with unique_id: {self.unique_id}>"

    def add_household(self, household: 'Household') -> None:
        """
        Adds household (not student) object to the list of the Neighbourhood
        object.

        Args:
            household (Household): Household object.
        """
        self.total += 1
        self.composition += household.attributes
        self.households.append(household)

    def remove_household(self, household: 'Household') -> None:
        """
        Removes household (not student) object from the list of the
        Neighbourhood object.

        Args:
            household (Household): Household object.
        """
        self.total -= 1
        self.composition -= household.attributes  # TODO: zero self.composition?
        self.households.remove(household)



"""
The Household class.
"""
import sys
from typing import List, ClassVar, Iterable
import numpy as np
from .agents_base import BaseAgent
from .school import School
from .student import Student
from .neighbourhood import Neighbourhood
from .functions import calc_comp_utility

MAX_INITIAL_HOUSEHOLDS = 100


class Household(BaseAgent):
    """
    The household object. Creates an expected number of students per
    household according to the student density, joins the closest
    neighbourhood.

    Args:
        unique_id (int): unique identifier of the agent.
        pos (tuple): (x,y) coordinates of the agent in the 2D-grid.
        model (CompassModel): CompassModel object.
        params (Argparser): containing all parameter values.
        category (int): the category [0,n-1] the agent belongs to.

    Attributes:
        unique_id (int): unique identifier of the agent.
        idx (int): index of this Household in lookup arrays
        pos (tuple): (x,y) coordinates of the agent in the 2D-grid.
        model (CompassModel): CompassModel object.
        params (Argparser): containing all (agent) parameter values.
        groups (list): containing all group types.
        attributes (array): array of attributes of the specific agent.
        composition (array): the sum of the attribute arrays of all Households
            in the local composition of this household.
        students (list): the student(s) in the household.
        school_id: the school.unique_id
        school: the school the first student attends
    """

    _total_households: ClassVar[int] = 0
    _max_households: ClassVar[int] = 0

    _household_res_utility: ClassVar[np.ndarray] = np.array([])
    _household_school_utility: ClassVar[np.ndarray] = np.array([])
    _household_distance: ClassVar[np.ndarray] = np.array([])
    _household_school_utility_comp: ClassVar[np.ndarray] = np.array([])
    _household_school_id: ClassVar[np.ndarray] = np.array([])

    __slots__ = ["idx", "category"]

    @classmethod
    def reset(cls, max_households: int = MAX_INITIAL_HOUSEHOLDS) -> None:
        """
        Allocates numpy arrays backing the Household data.

        Must be called before any Household objects are created.

        Arguments:
            max_households (int) : maximum number of Household objects that
            will be created.
        """
        if cls._total_households > 0:
            print(
                "Warning: resetting households while there already are "
                "Household objects."
            )
            cls._total_households = 0

        dtype = "float32"
        cls._max_households = max_households

        # was self.utility if residential
        cls._household_res_utility = np.zeros(max_households, dtype=dtype)

        # was self.utility if not residential
        cls._household_school_utility = np.zeros(max_households, dtype=dtype)

        cls._household_distance = np.zeros(max_households, dtype=dtype)
        cls._household_school_utility_comp = np.zeros(max_households, dtype=dtype)
        cls._household_school_id = np.zeros(max_households, dtype=dtype)

    def __init__(
        self,
        unique_id: int,
        pos: tuple[float, float],
        model: "CompassModel",
        params: dict,
        category: int,
        nhood: Neighbourhood = None,
    ):
        # Store parameters
        super().__init__(unique_id, pos, model, params)

        # Initialize some storage for Household objects
        # Note: this should be done by a call to Household.initialize()
        if Household._max_households == 0:
            print("Household.reset() not called yet, starting with default value")
            Household.reset()

        self.idx: int = Household._total_households
        Household._total_households += 1

        if Household._total_households > Household._max_households:
            print("Too many Household objects!")
            sys.exit(-1)

        self.category: int = category
        self.params: dict = params
        self.attributes: np.ndarray = self.attribute_array()
        self.composition: np.ndarray = self.new_composition_array()
        self.school_id: int = 0
        self.school: School = None

        # Create students
        self.students: List[Student] = []
        for i in range(int(self.params["student_density"])):
            self.students.append(Student(self.model.get_agents("amount"), self))

        # Join closest neighbourhood if applicable
        if self.params["n_neighbourhoods"]:

            # Join the given neighbourhood or else the closest
            if nhood:
                self.join_neighbourhood(nhood)
            else:
                neighbourhood = self.get_closest_neighbourhood(self.pos)
                self.join_neighbourhood(neighbourhood)

    def __repr__(self) -> str:
        """
        Returns:
            str: representing the unique identifier of the agent.
        """
        return f"<Household object with unique_id: {self.unique_id}>"

    @property
    def school_id(self) -> float:
        return Household._household_school_id[self.idx]

    @school_id.setter
    def school_id(self, value: float) -> None:
        Household._household_school_id[self.idx] = value

    @property
    def res_utility(self) -> float:
        return Household._household_res_utility[self.idx]

    @res_utility.setter
    def res_utility(self, value: float) -> None:
        Household._household_res_utility[self.idx] = value

    @property
    def school_utility(self) -> float:
        return Household._household_school_utility[self.idx]

    @school_utility.setter
    def school_utility(self, value: float) -> None:
        Household._household_school_utility[self.idx] = value

    @property
    def school_utility_comp(self) -> float:
        return Household._household_school_utility_comp[self.idx]

    @school_utility_comp.setter
    def school_utility_comp(self, value: float) -> None:
        Household._household_school_utility_comp[self.idx] = value

    @property
    def distance(self):
        return Household._household_distance[self.idx]

    @distance.setter
    def distance(self, value):
        Household._household_distance[self.idx] = value

    def attribute_array(self) -> np.ndarray:
        """
        This function creates the attribute array for the household that is
        used to calculate the local, neighbourhood and school compositions.
        """
        attributes = np.zeros(len(self.params["group_types"][0]))
        attributes[self.category] += 1
        return attributes

    def get_data(self, residential: bool) -> np.ndarray:
        """
        Gets the data of a specific agent for storing purposes in utils.py

        Args:
            residential (Bool): True if the model is in the residential process

        Notes:
            Only the school of the first student is used!
        """
        data = [
            self.pos[0],
            self.pos[1],
            self.composition[0],
            self.composition[1],
            self.res_utility if residential else self.school_utility,
            self.category,
            self.unique_id,
            self.distance,
            self.neighbourhood.unique_id
            if residential
            else self.students[0].school.unique_id,
        ]

        return np.array(data, dtype=float)

    def move_to_empty(
        self,
        empties: List[tuple[float, float]],
        num_considered: int,
        ranking_method: str,
    ) -> None:
        """
        Moves agent to a random empty cell, vacating the agent's old cell.

        Args:
            empties (list): a list of empty coordinates [(x1,y1),..,(xn,yn)]
            num_considered (int): how many spots are considered for the ranking
            ranking_method (str): one of 'highest' or 'proportional'
        """
        if len(empties) == 0:
            raise Exception("ERROR: No empty cells")

        # Pick possible empty locations, rank them and move to the chosen one
        idxs = self.random.choice(range(len(empties)), size=num_considered)
        positions = [empties[idx] for idx in idxs]
        new_pos = self.residential_ranking(positions, ranking_method)
        self.residential_move(old_pos=self.pos, new_pos=new_pos)

    def residential_move(
        self, old_pos: tuple[float, float] = None, new_pos: tuple[float, float] = None
    ) -> None:
        """
        Moves a household from old position to its new position.

        Args:
            old_pos (tuple): takes a tuple of integers (x,y), x<width, y<height.
            new_pos (tuple): takes a tuple of integers (x,y), x<width, y<height.
        """

        # Remove the agent from the old neighbourhood and add to the new one
        self.remove_neighbourhood(self.neighbourhood)
        self.model.grid.move_agent(self, new_pos)

        # Only remove the old position from the set if the agent actually
        # moves to a new position
        if new_pos != old_pos:
            self.model.grid.empties.discard(new_pos)
            self.model.grid.empties.add(old_pos)

        neighbourhood = self.get_closest_neighbourhood(self.pos)
        self.join_neighbourhood(neighbourhood)

        # Switch the attributes to the new location as well.
        self.model.switch_attrs(old_pos, new_pos)

    def step(
        self,
        residential: bool = False,
        initial_schools: bool = False,
        move_allowed: bool = True,
    ) -> int:
        """
        Steps the agent in the residential or school choice process.

        Args:
            residential (bool): equals True if the model needs to run a
                residential or a school step (default=False).
            initial_schools (bool): equals True if all schools are empty and
                students need an initial allocation first.
            move_allowed (True): equals True if the agent belongs to the
                percentage of agents that is allowed to move.

        Returns:
            int: boolean integer indicating if an agent was moved, to use in
                tracking of moved agents
        """

        # Run advancement for regular residential Schelling model
        if residential:

            # Check if move is allowed
            if not move_allowed:
                return 0

            # Check if there are neighbourhoods to choose from
            if self.params["n_neighbourhoods"] == 0:
                print("There are no neighbourhoods to choose from!")
                return 0

            elif self.params["household_density"] < 1:
                self.move_to_empty(
                    empties=list(self.model.grid.empties),
                    num_considered=self.params["num_considered"],
                    ranking_method=self.params["ranking_method"],
                )

            elif self.params["household_density"] == 1:
                print("Future place for switching of agents.")
                raise NotImplementedError

            return 1

        else:
            # Schools steps are done in the scheduler now for efficiency
            pass

    def update_residential(self) -> None:
        """
        Computes the composition and utility at the current residential location
        and sets the attributes to be used in other calculations.
        """

        category = self.category
        idx = self.idx
        if self.neighbourhood.total > 0:
            norm = 1.0 / self.neighbourhood.total
        else:
            norm = 1.0
        self.model.neighbourhood_compositions[idx] = (
            self.neighbourhood.composition[category] * norm
        )

        if self.params["neighbourhood_mixture"] == 1:
            # Only neighbourhood composition necessary, for case studies and
            # non-integer locations (then integer indexing is not possible)
            self.composition = self.neighbourhood.composition
        else:
            x, y = self.pos
            self.composition = self.model.compositions[x, y, :]
            self.model.local_compositions[idx] = self.model.normalized_compositions[
                x, y, :
            ][category]

    def residential_utility(
        self, composition: np.ndarray, neighbourhood_composition: np.ndarray = None
    ) -> float:
        """
        Compute residential utility.

        Args:
            composition (array): normalized local composition counts.
            neighbourhood_composition (array, optional): normalized
                neighbourhood composition counts.

        Returns:
            float: residential utility for a households' current location.

        Note:
            Ideally these computations are executed in numpy arrays,
            simultaneously for all agents, but this is not implemented for
            neighbourhoods yet.
        """
        params = self.params

        if neighbourhood_composition is not None:
            combined = (
                composition * (1 - params["neighbourhood_mixture"])
                + neighbourhood_composition * params["neighbourhood_mixture"]
            )
        else:
            combined = composition

        actual_fraction = combined[self.category]
        utility_at_max = params["utility_at_max"][0][self.category]
        optimal_fraction = params["optimal_fraction"][0][self.category]

        result = np.zeros_like(actual_fraction)
        calc_comp_utility(result, actual_fraction, utility_at_max, optimal_fraction)

        return result

    def get_closest_neighbourhood(self, pos: tuple[float, float]) -> Neighbourhood:
        """
        Find the closest neighbourhood object.

        Returns:
            neighbourhood: the closest (Euclidean) neighbourhood object.
        """
        location = self.model.closest_neighbourhoods[str(pos)]
        neighbourhood = self.model.location_to_agent[location]
        return neighbourhood

    def join_neighbourhood(self, neighbourhood: Neighbourhood) -> None:
        """
        Join the given neighbourhood object.

        Args:
            neighbourhood (Neighbourhood): a neighbourhood object.
        """
        self.neighbourhood = neighbourhood
        neighbourhood.add_household(self)

    def get_neighbourhood(self):
        """
        Return the neighbourhood this household belongs to.
        """
        return self.neighbourhood

    def remove_neighbourhood(self, neighbourhood: Neighbourhood) -> None:
        """
        Leave a neighbourhood.

        Args:
            neighbourhood (Neighbourhood): a neighbourhood object.
        """
        neighbourhood.remove_household(self)
        self.neighbourhood = None

    def school_ranking_initial(self) -> Iterable[School]:
        """
        Computes a list containing all schools ranked to preference. The initial
        school ranking is random.

        Note: when the number of steps is small, this function shows up as a bottleneck.
        the choice() function is very slow, probably because to get sufficient
        randomness it makes many calls to a real random number generator.

        At the moment, there are two implementations, one is faster for the test case,
        the other one for much larger cases.

        Returns:
            list: a randomly ordered list of School objects.
        """
        schools = self.model.get_agents("schools")

        if True:
            # create a list of indices [0 .. n_schools-1]
            idxs = list(range(len(schools)))

            # shuffle it in-place
            self.model.random.shuffle(idxs)

            # create a list with the shuffled school objects
            return [schools[idx] for idx in idxs]

        if False:
            schools = schools.copy()
            self.model.random.shuffle(schools)
            return schools

    def residential_ranking(
        self, positions: List[tuple[float, float]], ranking_method: str
    ) -> tuple[float, float]:
        """
        Computes the ranked location prefences of a household.

        Args:
            positions (list): list of (x, y) tuples that are considered.
            ranking_method (str): one of 'highest' or 'proportional'

        Returns:
            tuple: new position (x, y) of the household.
        """

        summed = 0
        max_utility = 0
        params = self.params
        positions = list(positions) + [self.pos]  # Append own position
        utilities = np.zeros(len(positions))
        temperature = params["temperature"]
        compositions = self.model.compositions
        norm_compositions = self.model.normalized_compositions

        for index, pos in enumerate(positions):

            if pos == self.pos:
                utility = self.res_utility
            else:
                #  ASSUMING AGENTS HAVE THE SAME RADIUS HERE
                x, y = pos
                composition = compositions[x, y, :]
                norm_composition = norm_compositions[x, y, :]
                neighbourhood = self.get_closest_neighbourhood(pos)
                if neighbourhood.total > 0:
                    norm = 1.0 / neighbourhood.total
                else:
                    norm = 1.0
                utility = self.residential_utility(
                    norm_composition, neighbourhood.composition * norm
                )

            if utility >= max_utility:
                max_utility = utility
                new_pos = [pos]

            utility = np.exp(temperature * utility)
            summed += utility
            utilities[index] = utility

        if summed == float("inf"):
            # we're having overflow issues
            # FIXME: just create some reasonable weights
            top = np.argmax(utilities)
            utilities[:] = 0.0
            utilities[top] = 1.0
        else:
            utilities = np.nan_to_num(utilities / summed, copy=False)
        if ranking_method == "proportional" or ranking_method:
            idxs = self.random.choice(range(len(positions)), p=utilities, size=1)
            new_pos = [positions[idxs[0]]]

        return new_pos[0]

    def get_student_count(self) -> int:
        """
        Calculates the number of students in a household.

        Returns:
            int: the amount of students in the household.
        """
        return len(self.students)

    def get_shock(self) -> float:
        """ " Returns a small random float value"""
        return self.model.get_shock()

    def get_uniform_shock(self) -> float:
        """
        Returns a random value between 0 and 1.
        """
        return self.model.get_uniform_shock()



#!/usr/bin/env python

from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="compass-school",
    version="v1.0.1",
    description="This Python implementation tries to model school choice and resulting school segregation based on the work of Schelling (1971) and Stoica & Flache (2014).",
    author="Eric Digum, Jisk Attenma, Ji Qi",
    author_email="e.p.n.dignum@uva.nl, j.attema@esciencecenter.nl, j.qi@esciencecenter.nl",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=["compass"],
    python_requires=">=3.10",
    url="https://github.com/ODISSEI-School-Choice/school-choice",
    download_url="https://github.com/ODISSEI-School-Choice/school-choice/archive/refs/tags/v1.0.0.tar.gz",
    install_requires=[
        "bokeh",
        "hypothesis",
        "ijson",
        "Mesa",
        "pytest",
        "scikit_learn",
        "scipy",
        "seaborn",
        "dask",
        "distributed",
        "geopandas",
        "shapely"
    ],
)



"""
The time handler. The scheduler defined in this file is equivalent to the
SimultaneousActivation scheduler from MESA v0.8.6, with the addition of
randomization within the step and advance phase plus the possibility of
parallelization over cores. The execution of the step() and advance() parts in
the agents used, result in a system that is not entirely simultaneous.
"""

from collections import OrderedDict
import numpy as np
from .allocator import Allocator


class ThreeStagedActivation:
    """
    A scheduler to simulate the simultaneous activation of all the agents.

    Args:
        model (CompassModel): CompassModel object.

    Attributes:
        model (CompassModel): CompassModel object.
        steps (int): counts the number of steps.
        time (int): counts the number of time steps.
        moved_agents (int): amount of agents moved in a time step.
        max_movements (int): maximum amount of agents to be moved per time step.
        _agents (dict): contains all the agents to be moved.
        allocater (Allocater): an Allocater object.

    Todo:
        * Parallelization if needed.
    """

    def __init__(self, model):

        self.model = model
        self.params = model.params
        (
            self.time,
            self.split_index,
            self.moved_agents,
            self.school_steps,
            self.residential_steps,
        ) = (0, 0, 0, 0, 0)

        # A fraction of 0.40 means three splits as does 0.35 for example.
        self.n_splits = np.ceil(1 / self.params["max_move_fraction"])
        self.step_size = 1 + int(self.model.params["n_households"] // self.n_splits)
        self.max_movements = 1 + int(
            self.model.params["max_move_fraction"] * self.model.params["n_students"]
        )
        self._agents = dict()
        self.allocator = Allocator()

    def add(self, agent):
        """
        Add an Agent object to the schedule.

        Args:
            agent: an Agent to be added to the schedule.

        Note:
            The agent must have a step() method.
        """
        if agent.__class__.__name__ not in self._agents:
            self._agents[agent.__class__.__name__] = OrderedDict()
        self._agents[agent.__class__.__name__][agent.unique_id] = agent

    def remove(self, agent):
        """
        Remove all instances of a given agent from the schedule.

        Args:
            agent: an Agent object.
        """
        del self._agents[agent.__class__.__name__][agent.unique_id]

    def get_agent_count(self):
        """
        Returns:
            int: the current number of agents in the queue.
        """
        return sum([len(agents) for agents in self._agents.values()])

    def get_time(self, time_type="time"):
        """
        Returns:
            int: current time step (total, residential or school)
        """

        if time_type == "time":
            return self.time
        if time_type == "residential":
            return self.residential_steps
        if time_type == "school":
            return self.school_steps

        return None

    def agents_to_move(self, households, initial_schools):
        """
        Determine which households need to be moved this step.
        """

        # Necessary because of the reset button in the visualisation, it's too
        # early in the init()
        if self.get_time() == 0:
            # A fraction of 0.40 means three splits as does 0.35 for example.
            self.n_splits = np.ceil(1 / self.params["max_move_fraction"])
            self.step_size = 1 + int(self.model.params["n_households"] // self.n_splits)
            self.max_movements = 1 + int(
                self.model.params["max_move_fraction"] * self.model.params["n_students"]
            )

        if self.params["scheduling"] == 0 or initial_schools:
            self.model.random.shuffle(households)
            self.lower_index = 0
            self.upper_index = -1

        # Determine the new splits as everyone has been stepped
        elif self.split_index % self.n_splits == 0:
            self.model.random.shuffle(households)
            self.split_index = 0
            self.lower_index = 0
            self.upper_index = self.step_size
            self.split_index += 1

        else:
            self.lower_index = self.upper_index
            self.upper_index = (1 + self.split_index) * self.step_size
            self.split_index += 1

        return households[self.lower_index : self.upper_index]

    def step(self, residential=False, initial_schools=False):
        """
        Steps all agents and then advances them.

        Args:
            residential (bool): True if a residential step needs to be
                performed, False (default) means a school step.
            initial_schools (bool): True if an initial school step needs to be
                performed, False (default) means a school step.
        """

        all_households = self.model.get_agents("households")
        households_to_move = self.agents_to_move(all_households, initial_schools)

        # FIXME: we shouldnt store this, but is needed for a test
        self.households_to_move = households_to_move  # For testing purposes

        if residential:

            # Rankings are still calculated in the Household object instead of
            # model wide as for the schools.
            for household in households_to_move:
                household.step(residential=residential, initial_schools=initial_schools)

            # UPDATE COMPOSITIONS OF ALL AGENTS AFTER ALL THE MOVES
            self.model.calc_residential_compositions()
            for household in all_households:
                household.update_residential()
            self.model.calc_res_utilities()
            self.residential_steps += 1

        else:

            if initial_schools:

                # Initial allocation for EVERY household
                # Set initial preferences
                if self.params["case"].lower() == "lattice":
                    self.model.compute_school_distances()  # execute only in lattice case
                    p = self.params["p"]
                    q = self.params["q"]
                    self.model.distance_utilities = 1.0 / (
                        1 + (self.model.all_distances / p) ** q
                    )

                #TODO: remove commented code
                #      initialization happens in new allocator
                #for household in all_households:
                #    for student in household.students:
                #        student.set_school_preference(
                #            household.school_ranking_initial()
                #        )

                # Allocate students after all initial preferences have been set
                self.allocate_schools(all_households, initial_schools)

            else:
                # Normal school step
                self.model.calc_school_rankings(
                    households_to_move, self.model.get_agents("schools")
                )
                self.allocate_schools(households_to_move, initial_schools)

            # Calculate the new school compositions:
            #  * Compass.school_compositions
            #  TODO: should this move to in the 'residential' below?
            self.model.calc_school_compositions()

            if residential:
                # Set:
                #  * Compass.neighbourhodd_compositions
                #  * Compass.local_compositions
                #  * Household.composition
                for household in all_households:
                    household.update_residential()
            # updating the household.distance is done when moving the student

            # Set:
            #  * Household.school_utility_comp
            #  * Household.school_utility
            self.model.calc_school_utilities()
            self.school_steps += 1

        # Updating of the utilities of ALL agents was done in
        # self.model.calc_school_utilities()

        self.time += 1
        self.model.measurements.end_step(residential)

    def allocate_schools(self, households, initial_schools):
        """
        Allocates students to schools.

        Args:
            households (list): list of all the households that need to be
                allocated.
            initial_schools (bool): True if an initial school step needs to be
                performed, False (default) means a school step.
        """
        if initial_schools:
            self.allocator.initial_school(households)
        else:
            self.allocator.optimal_school(households)
